==> lein-template-0.1.4.jar: leiningen/new/smart_function/core.clj <==
(ns {{sanitized}}.core
  (:refer-clojure :exclude [max min get inc dec + - * / quot mod == rem contains? get-in < <= > >=
                            boolean re-find and or count str nth rand nil? hash-set empty? not])
  (:require
    [{{sanitized}}.internal :as internal]
    [{{sanitized}}.custom_functions :refer :all]
    [clojure.string :as string]
    [cheshire.core :as json]
    [clojure.tools.logging :as log]
    [aleph.http :as http])
  (:import (java.time Instant)))


(defn resolve-two-tuple
  [subject]
  (-> (internal/query-send 0 1000 {:select "?sid"
                                   :where  [["?sid" (first subject) (second subject)]]}) first))

(defn resolve-cname
  [_id]
  (if    (number? _id)
    (let [collection-sid (-> (bit-shift-right _id 44)
                             (bit-or 17592186044416))
          cname-query {:select "?cName"
                       :where [[collection-sid "_collection/name" "?cName"]]}
          res       (-> (internal/query-send 0 1000 cname-query) first)]
      res)
    (clojure.core/re-find #"^[a-zA-Z0-9_][a-zA-Z0-9\.\-_]{0,254}" _id)))

(defn get-sid
  [subject]
  (cond
    (number? subject)
    subject

    (string? subject)
    nil

    (clojure.core/and (vector? subject) (= 2 (clojure.core/count subject)))
    (resolve-two-tuple subject)))

(defn get-pid
  [predicate-name]
  (if (clojure.core/nil? predicate-name)
    nil
    (-> (internal/query-send 0 1000 {:select "?id" :where  [["?id" "_predicate/name" predicate-name]]})
        first)))


(defn create-ctx-transact
  [authId tx pName subject]
  {:instant (Instant/now)
   :sid     (-> (:_id tx) get-sid)
   :pid     (get-pid pName)
   :auth_id authId
   :state   (atom {:stack   []
                   :credits 0
                   :spent   0})})

(defn keyword->str
  "Converts a keyword to string. Can safely be called on a
  string which will return itself."
  [k]
  (cond
    (keyword? k) (subs (clojure.core/str k) 1)
    (string? k) k
    :else (throw (ex-info (clojure.core/str "Cannot convert type " (type k) " to string: " (pr-str k))
                          {:status 500 :error :db/unexpected-error}))))

(defn namespace-predicate
  [collection predicate]
  (if (.contains predicate "/")
    predicate
    (clojure.core/str collection "/" predicate)))

(defn get-predicate+tx-specs
  [db predicate]
  (-> (internal/query-send 0 1000 {:select {"?predicate" [{"_predicate/spec" ["*"]} {"_predicate/txSpec" ["*"]}]}
                                   :where [["?predicate" "_predicate/name" predicate]]}) first))

(defn create-ctx-predicate-spec
  [db authId subject pName sid collection]
  {:instant (Instant/now)
   :sid     (get-sid subject)
   :pid     (-> (resolve-cname sid)
                (namespace-predicate (keyword->str pName))
                get-pid)
   :auth_id authId
   :state   (atom {:stack   []
                   :credits 0
                   :spent   0})})



;; TO DO - Filter rules properly. Everything else working as expected
(defn find-all-applicable-rules
  [db authId tx]
  ;; Find collection spec
  ;; Find predicate and predicate tx spec
  ;; Find all auth -> rules
  (let [_id         (:_id tx)
        collection  (if (clojure.core/and (vector? _id) (= 2 (clojure.core/count _id)))
                      (->> (resolve-two-tuple _id)
                           resolve-cname)
                      (resolve-cname _id))
        collection-spec (-> (internal/query-send 0 1000 {:select {"?collection" [{"_collection/spec" ["*"]}]}
                                                         :where [["?collection" "_collection/name" collection]]}) first)
        all-predicates (->> (dissoc tx :_id) keys
                            (mapv keyword->str) (map #(namespace-predicate collection %)))
        predicate+tx-specs (mapv (fn [pred]
                                   (let [predicate+tx-specs (get-predicate+tx-specs db pred)]
                                     (if-not (clojure.core/nil? predicate+tx-specs)
                                       {:name   pred
                                        :spec   (:_predicate/spec predicate+tx-specs)
                                        :txSpec (:_predicate/txSpec predicate+tx-specs)}))) all-predicates)
        predicate+tx-specs'  (filterv #(clojure.core/not (clojure.core/nil? %)) predicate+tx-specs)
        auth-roles    (-> (internal/query-send 0 1000 {:select [{:_auth/roles [{"_role/rules" ["*"]}]}]
                                                       :from authId}) :_auth/roles)]
    {:_collection {:name collection
                   :spec (:_collection/spec collection-spec)}
     :_predicate predicate+tx-specs'
     :_auth {:_id authId
             :rules auth-roles}}))

(defn replace-fn
  [funStr matches replacement shift]
  (reduce (fn [funCnt match]
            (let [[fun last-idx] funCnt
                  startMatch (string/index-of fun match last-idx)
                  length     (clojure.core/count match)
                  endMatch   (clojure.core/- (clojure.core/+ startMatch length) shift)
                  funStr'    (clojure.core/str (subs fun 0 endMatch) replacement (subs fun endMatch))]
              [funStr' endMatch]))
          [funStr 0] matches))

(defn filter-vals
  "Filters map k/v pairs dropping any where predicate applied to value is false."
  [pred m]
  (reduce-kv (fn [m k v] (if (pred v) (assoc m k v) m)) {} m))

(defn without-nils
  "Remove all keys from a map that have nil or empty collection values."
  [m]
  (filter-vals #(if (coll? %) (not-empty %) (some? %)) m))

(defn insert-ctx-param
  [funStr]
  (let [matches         (re-seq #"\([a-zA-Z-\?\<\>=\_]+ " funStr)
        [funStr']       (replace-fn funStr matches "ctx " 0)
        matches'        (re-seq #"\([a-zA-Z-\?\<\>=\_]+\)" funStr')
        [funStr'']      (replace-fn funStr' matches' " ctx" 1)] funStr''))

(defn gen-flakes
  [txn]
  (let [{:keys [ip network db]} (internal/config)
        res   (-> @(http/post (clojure.core/str ip "/fdb/" network "/" db "/gen-flakes")
                              {:content-type    :json
                               :request-timeout 5000                 ;; 5 seconds
                               :body            (json/encode txn)})
                  :body
                  internal/json-read)]
    (:flakes res)))

(defn gen-ctxs
  [flakes opts]
  (let [{:keys [type auth_id predicate collection]} opts
        instant (Instant/now)
        pid     (get-pid predicate)
        _ (when (clojure.core/and (clojure.core/nil? pid)  (#{"_predicate/spec"} type))
            (throw (ex-info "When testing a _predicate/spec, you must provide the predicate name."
                            {:status 400
                             :error :db/invalid-txn})))
        test-iterators (condp = type "_predicate/spec"
                                     (let [flakes'  (filterv #(= pid (second %)) flakes)
                                           sids (map first flakes')]
                                       (map #(assoc {} :sid %) sids)))
        ctx-base (-> {:instant instant
                      :flakes flakes
                      :auth_id auth_id
                      :pid pid} without-nils)]
    (map #(merge ctx-base %) test-iterators)))

(comment

  (def flakes (gen-flakes [{:_id "_user" :username "anna"}]))
  (gen-ctxs flakes {:type "_predicate/spec" :predicate "_user/username"})

  flakes
  )


(defn test-fn
  [txn spec opts]
  (let [{:keys [type auth_id params]} opts
        flakes    (gen-flakes txn)
        ctxs      (gen-ctxs flakes opts)
        param-sets (map #(concat [%] params) ctxs)]
    (map #(apply spec %) param-sets)))



;     function-string (insert-ctx-param smart-fun-string)
;     fn-str  (read-string function-string)
;     custom-func (list #'clojure.core/fn ['ctx] fn-str)
;     func'        (cons custom-func ['ctx])
;     f               `(fn [~'ctx] ~func')
;     f'   (eval f)]
; (f' ctx)))



(comment
  ; Let's say I want to test "required".
  ; I think my required function returns true if the predicate is there. False if not.
  ; But I want to test it out.
  ; (test-fn [{:_id "_user" :username "anna"}]
  ; required {:params ["_user/username"]  :type "_predicate/spec"})
  ;
  ; 1. Function issues gen-flakes
  ; 1a. If the txn isn't valid (before adding the leiningen spec), it throws an error
  ; 1b. If the txn IS valid, it returns flakes

  ; 2. Now, we create a ctx, pred-spec style.
  ; That sid or subject is now a real subject, even if we were creating a leiningen subject originally

  ; 3. Then, we can run the function against the context.
  ; We take flakes from ctx -> and if we need to do any queries, it's a query-with
  ; sid/ pid, etc everything else that may have had to be updated was updated in ctx.

  ; 4. Returns the result of the function, complete with stack.

  (test-fn [{:_id "_user" :username "anna"}]
           required {:params ["_user/username"]  :type "_predicate/spec" :predicate "_user/username"})

  ; (test-fn [{:_id ["_user/username"]}])


  ; (defn required
  ;   [ctx predicate]
  ;   (boolean ctx (get ctx (?s ctx) predicate)))
  )

==> lein-template-0.1.5.jar: leiningen/new/smart_function/core.clj <==
(ns {{sanitized}}.core
  (:refer-clojure :exclude [max min get inc dec + - * / quot mod == rem contains? get-in < <= > >=
                            boolean re-find and or count str nth rand nil? hash-set empty? not])
  (:require
    [{{sanitized}}.internal :as internal]
    [{{sanitized}}.custom_functions :refer :all]
    [{{sanitized}}.temp_custom_functions :refer :all]
    [clojure.string :as string]
    [cheshire.core :as json]
    [clojure.tools.logging :as log]
    [aleph.http :as http])
  (:import (java.time Instant)))

(defn resolve-two-tuple
  [subject]
  (-> (internal/query-send 0 1000 {:select "?sid"
                                   :where  [["?sid" (first subject) (second subject)]]}) first))

(defn resolve-cname
  [_id]
  (if    (number? _id)
    (let [collection-sid (-> (bit-shift-right _id 44)
                             (bit-or 17592186044416))
          cname-query {:select "?cName"
                       :where [[collection-sid "_collection/name" "?cName"]]}
          res       (-> (internal/query-send 0 1000 cname-query) first)]
      res)
    (clojure.core/re-find #"^[a-zA-Z0-9_][a-zA-Z0-9\.\-_]{0,254}" _id)))

(defn get-sid
  [subject]
  (cond
    (number? subject)
    subject

    (string? subject)
    nil

    (clojure.core/and (vector? subject) (= 2 (clojure.core/count subject)))
    (resolve-two-tuple subject)))

(defn get-pid
  [predicate-name]
  (if (clojure.core/nil? predicate-name)
    nil
    (-> (internal/query-send 0 1000 {:select "?id" :where  [["?id" "_predicate/name" predicate-name]]})
        first)))


(defn create-ctx-transact
  [authId tx pName subject]
  {:instant (Instant/now)
   :sid     (-> (:_id tx) get-sid)
   :pid     (get-pid pName)
   :auth_id authId
   :state   (atom {:stack   []
                   :credits 0
                   :spent   0})})

(defn keyword->str
  "Converts a keyword to string. Can safely be called on a
  string which will return itself."
  [k]
  (cond
    (keyword? k) (subs (clojure.core/str k) 1)
    (string? k) k
    :else (throw (ex-info (clojure.core/str "Cannot convert type " (type k) " to string: " (pr-str k))
                          {:status 500 :error :db/unexpected-error}))))

(defn namespace-predicate
  [collection predicate]
  (if (.contains predicate "/")
    predicate
    (clojure.core/str collection "/" predicate)))

(defn get-predicate+tx-specs
  [db predicate]
  (-> (internal/query-send 0 1000 {:select {"?predicate" [{"_predicate/spec" ["*"]} {"_predicate/txSpec" ["*"]}]}
                                   :where [["?predicate" "_predicate/name" predicate]]}) first))

(defn create-ctx-predicate-spec
  [db authId subject pName sid collection]
  {:instant (Instant/now)
   :sid     (get-sid subject)
   :pid     (-> (resolve-cname sid)
                (namespace-predicate (keyword->str pName))
                get-pid)
   :auth_id authId
   :state   (atom {:stack   []
                   :credits 0
                   :spent   0})})



;; TO DO - Filter rules properly. Everything else working as expected
(defn find-all-applicable-rules
  [db authId tx]
  ;; Find collection spec
  ;; Find predicate and predicate tx spec
  ;; Find all auth -> rules
  (let [_id         (:_id tx)
        collection  (if (clojure.core/and (vector? _id) (= 2 (clojure.core/count _id)))
                      (->> (resolve-two-tuple _id)
                           resolve-cname)
                      (resolve-cname _id))
        collection-spec (-> (internal/query-send 0 1000 {:select {"?collection" [{"_collection/spec" ["*"]}]}
                                                         :where [["?collection" "_collection/name" collection]]}) first)
        all-predicates (->> (dissoc tx :_id) keys
                            (mapv keyword->str) (map #(namespace-predicate collection %)))
        predicate+tx-specs (mapv (fn [pred]
                                   (let [predicate+tx-specs (get-predicate+tx-specs db pred)]
                                     (if-not (clojure.core/nil? predicate+tx-specs)
                                       {:name   pred
                                        :spec   (:_predicate/spec predicate+tx-specs)
                                        :txSpec (:_predicate/txSpec predicate+tx-specs)}))) all-predicates)
        predicate+tx-specs'  (filterv #(clojure.core/not (clojure.core/nil? %)) predicate+tx-specs)
        auth-roles    (-> (internal/query-send 0 1000 {:select [{:_auth/roles [{"_role/rules" ["*"]}]}]
                                                       :from authId}) :_auth/roles)]
    {:_collection {:name collection
                   :spec (:_collection/spec collection-spec)}
     :_predicate predicate+tx-specs'
     :_auth {:_id authId
             :rules auth-roles}}))

(defn replace-fn
  [funStr matches replacement shift]
  (reduce (fn [funCnt match]
            (let [[fun last-idx] funCnt
                  startMatch (string/index-of fun match last-idx)
                  length     (clojure.core/count match)
                  endMatch   (clojure.core/- (clojure.core/+ startMatch length) shift)
                  funStr'    (clojure.core/str (subs fun 0 endMatch) replacement (subs fun endMatch))]
              [funStr' endMatch]))
          [funStr 0] matches))

(defn filter-vals
  "Filters map k/v pairs dropping any where predicate applied to value is false."
  [pred m]
  (reduce-kv (fn [m k v] (if (pred v) (assoc m k v) m)) {} m))

(defn without-nils
  "Remove all keys from a map that have nil or empty collection values."
  [m]
  (filter-vals #(if (coll? %) (not-empty %) (some? %)) m))

(defn insert-ctx-param
  [funStr]
  (let [matches         (re-seq #"\([a-zA-Z-\?\<\>=\_]+ " funStr)
        [funStr']       (replace-fn funStr matches "ctx " 0)
        matches'        (re-seq #"\([a-zA-Z-\?\<\>=\_]+\)" funStr')
        [funStr'']      (replace-fn funStr' matches' " ctx" 1)] funStr''))

(defn gen-flakes
  [txn]
  (let [{:keys [ip network db]} (internal/config)]
    (try (-> @(http/post (clojure.core/str ip "/fdb/" network "/" db "/gen-flakes")
                         {:content-type    :json
                          :request-timeout 5000                 ;; 5 seconds
                          :body            (json/encode txn)})
             :body
             internal/json-read
             :flakes)
         (catch Exception e (clojure.core/str "Error in gen-flakes: " (-> (ex-data e)
                                                                          :body
                                                                          internal/json-read))))))

(defn gen-ctxs
  [flakes opts]
  (let [{:keys [type auth_id predicate collection]} opts
        instant (Instant/now)
        pid     (get-pid predicate)
        _ (when (clojure.core/and (clojure.core/nil? pid)  (#{"_predicate/spec"} type))
            (throw (ex-info "When testing a _predicate/spec, you must provide the predicate name."
                            {:status 400
                             :error :db/invalid-txn})))
        _ (when (clojure.core/and (clojure.core/nil? collection)  (#{"_collection/spec"} type))
            (throw (ex-info "When testing a _collection/spec, you must provide the collection name."
                            {:status 400
                             :error :db/invalid-txn})))
        test-iterators (condp = type
                         "_predicate/spec"
                         (let [flakes'  (filterv #(= pid (second %)) flakes)]
                           (map (fn [flake]
                                  {:sid (first flake)
                                   :o (clojure.core/nth flake 2)}) flakes'))

                         "_collection/spec"
                         (let [flakes'  (filterv #(= collection (resolve-cname (first %))) flakes)]
                           (map (fn [flake]
                                  {:sid (first flake)}) flakes')))
        ctx-base (-> {:instant instant
                      :flakes flakes
                      :auth_id auth_id
                      :pid pid
                      :print-stack true
                      :state   (atom {:stack   []
                                      :credits 10000000
                                      :spent   0})} without-nils)]
    (map #(merge ctx-base %) test-iterators)))

(defn test-fn
  [txn spec opts]
  (let [{:keys [type auth_id params]} opts
        flakes    (gen-flakes txn)
        _         (println "FLAKES" flakes)
        ctxs      (gen-ctxs flakes opts)
        _         (log/info "CTXS" ctxs)
        param-sets (map #(concat [%] params) ctxs)]
    (map #(apply spec %) param-sets)))
